@@ -1,7 +1,12 @@
+## -*- cperl -*-
+
 package XML::DT;
 
 BEGIN{
- use XML::Parser;
+
+ use XML::LibXML;
+#XML::PARSER# use XML::Parser;
+
  use Data::Dumper;
  use Exporter ();
  use vars qw($c %v $q @dtcontext %dtcontextcount @dtatributes );
@@ -10,7 +15,7 @@ BEGIN{
  @ISA=qw(Exporter);
  @EXPORT=qw(dt dtstring inctxt ctxt mkdtskel mkdtdskel toxml MMAPON $c %v $q 
          @dtcontext %dtcontextcount @dtatributes pathdt pathdtstring );
- $VERSION = '0.20';
+ $VERSION = '0.22';
 }
 
 =head1 NAME
@@ -41,11 +46,14 @@ XML::DT - a package for down translation of XML to strings
     ctxt(1)       /* the father element */
 
     mkdtskel($file)
+    mkdtskel("-latin1", $file)
     mkdtdskel($file)
 
 =head1 DESCRIPTION
 
-This module processes XML files with an approach similar to OMNIMARK.
+This module processes XML files with an approach similar to OMNIMARK. As 
+XML parser it uses XML::Parser or XML::LibXML module in an indenpendent way. At
+configure stage, you should choose one of the backends.
 
 Down translation function C<dt> receives a filename and a set of expressions
 (functions) defining the processing and associated values for each element.
@@ -298,27 +306,47 @@ sub dt {
   %ty=(%{$xml{'-type'}}) if defined($xml{'-type'}); 
   $ty{-ROOT} = "NONE";
 
-  # create a new XML::Parser instance using Tree Style
-  if (defined($xml{-inputenc}) && ($xml{-inputenc} eq 'ISO-8859-1')){
-     $parser = new XML::Parser(Style => 'Tree',
-                               ErrorContext => 2 ,
-                               ProtocolEncoding => 'ISO-8859-1');
-  }
-  else { $parser = new XML::Parser(Style => 'Tree',
-                                   ErrorContext => 2 ,
-                                  );
-  }
-
-  # execute Begin action if it exists
   if ($xml{-begin}){ &{$xml{-begin}} }
 
-  # Convert XML to Perl code
-  $tree = $parser->parsefile($file);
+# TODO --- how to force encoding with XML::LibXML?
+# $xml{-inputenc}
+
+# create a new LibXML parser
+ my $parser = XML::LibXML->new();
+
+# Check if we should expand entities
+  $parser->expand_entities(1) if defined $xml{'-noexpand'} && $xml{'-noexpand'};
+
+# parse the file
+  my $doc = $parser->parse_file($file);
+
+# get the document root element
+  my $tree = $doc->getDocumentElement();
+
+#XML::PARSER## create a new XML::Parser instance using Tree Style
+#XML::PARSER#  if (defined($xml{-inputenc}) && ($xml{-inputenc} eq 'ISO-8859-1')){
+#XML::PARSER#     $parser = new XML::Parser(Style => 'Tree',
+#XML::PARSER#                               ErrorContext => 2 ,
+#XML::PARSER#                               ProtocolEncoding => 'ISO-8859-1');
+#XML::PARSER#  }
+#XML::PARSER#  else { $parser = new XML::Parser(Style => 'Tree',
+#XML::PARSER#                                   ErrorContext => 2 ,
+#XML::PARSER#                                  );
+#XML::PARSER#  }
+
+#XML::PARSER## Convert XML to Perl code
+#XML::PARSER#  $tree = $parser->parsefile($file);
+
 
   # execute End action if it exists
-  if($xml{-end}){ $c= omni("-ROOT",\%xml,@$tree); 
-                  &{$xml{-end}} }
-  else          { omni("-ROOT",\%xml,@$tree) }
+  if($xml{-end}) {
+    $c= omni("-ROOT", \%xml, $tree);
+#XML::PARSER#    $c= omni("-ROOT", \%xml, @$tree);
+    &{$xml{-end}}
+  } else {
+    omni("-ROOT",\%xml, $tree)
+#XML::PARSER#    omni("-ROOT",\%xml, @$tree)
+  }
 }
 
 sub ctxt {
@@ -343,28 +371,52 @@ sub pathdtstring{
 sub dtstring
 { my ($string,%xml)=@_;
 
+  $xml{'-type'} = {} unless defined $xml{'-type'};
   %ty=(%{$xml{'-type'}}, -ROOT => "NONE");
 
-  # create a new XML::Parser instance using Tree Style
-  if (defined($xml{-inputenc}) && ($xml{-inputenc} eq 'ISO-8859-1')){
-     $parser = new XML::Parser(Style => 'Tree',
-                               ErrorContext => 2 ,
-                               ProtocolEncoding => 'ISO-8859-1');
-  }
-  else { $parser = new XML::Parser(Style => 'Tree',
-                                   ErrorContext => 2 ,
-                                  );
-  }
-
-  #execute Begin action if she exists
+  # execute Begin action if she exists
   if ($xml{-begin}){ &{$xml{-begin}} }
 
-  # Convert XML to Perl code (Tree)
-  $tree = $parser->parse($string);
+# TODO --- how to force encoding with XML::LibXML?
+# $xml{-inputenc}
+ $string = XML::LibXML::encodeToUTF8($xml{-inputenc},$string) if ($xml{-inputenc});
 
-  if($xml{-end}){ $c= omni("-ROOT",\%xml,@$tree); 
-                  &{$xml{-end}} }
-  else          { omni("-ROOT",\%xml,@$tree) }
+# create a new LibXML parser
+  my $parser = XML::LibXML->new();
+
+# Check if we should expand entities
+  $parser->expand_entities(0) if defined $xml{'-noexpand'} && $xml{'-noexpand'};
+
+# parse the string
+  my $doc = $parser->parse_string($string);
+
+# get the document root element
+  my $tree = $doc->getDocumentElement();
+
+#XML::PARSER## create a new XML::Parser instance using Tree Style
+#XML::PARSER#  if (defined($xml{-inputenc}) && ($xml{-inputenc} eq 'ISO-8859-1')){
+#XML::PARSER#     $parser = new XML::Parser(Style => 'Tree',
+#XML::PARSER#                               ErrorContext => 2 ,
+#XML::PARSER#                               ProtocolEncoding => 'ISO-8859-1');
+#XML::PARSER#  }
+#XML::PARSER#  else { $parser = new XML::Parser(Style => 'Tree',
+#XML::PARSER#                                   ErrorContext => 2 ,
+#XML::PARSER#                                  );
+#XML::PARSER#  }
+
+#XML::PARSER## Convert XML to Perl code (Tree)
+#XML::PARSER#  $tree = $parser->parse($string);
+
+
+  # Check if we have a end function
+  if ($xml{-end}) {
+    $c = omni("-ROOT", \%xml, $tree);
+#XML::PARSER#    $c = omni("-ROOT", \%xml, @$tree);
+    &{$xml{-end}}
+  } else { 
+    omni("-ROOT", \%xml, $tree)
+#XML::PARSER#    omni("-ROOT", \%xml, @$tree)
+  }
 }
 
 sub pathdt{
@@ -494,13 +546,15 @@ sub pathtodt {
 }
 
 sub omni{
-  my ($par,$xml,@l) = @_;
+  my ($par, $xml, @l) = @_;
   my $type = $ty{$par} || "STR";
-  my %typeargs=();
-  if(ref($type) eq "mmapon"){ 
-       for(@$type){$typeargs{$_}=1;}
-       $type = "MMAPON";
-       };
+  my %typeargs = ();
+
+  if (ref($type) eq "mmapon") {
+    for (@$type)
+      { $typeargs{$_} = 1 }
+    $type = "MMAPON";
+  };
 
   my $r ;
 
@@ -513,59 +567,88 @@ sub omni{
   elsif( $type eq 'NONE')                             { $r = "" }
   elsif( $type eq 'ZERO')                             { return "" }
 
-  my ($name, $val, @val,$atr, $aux);
-  while( @l) {
-    ($name, $val, @l) = @l;
-    if ($name eq "0"){ 
-      $name="-pcdata";
+  my ($name, $val, @val, $atr, $aux);
+
+  while(@l) {
+    my $tree = shift @l;
+    $name = $tree->getName();
+#XML::PARSER#    ($name,$val,@l) = @l;
+
+    if (ref($tree) eq "XML::LibXML::Comment") {
+
+      ### At the moment, treat as Text
+      ### We will need to change this, I hope!
+      $val = "";
+      $name = "-pcdata";
       $aux= (defined($xml->{-outputenc}) && $xml->{-outputenc} eq 'ISO-8859-1')
-              ?lat1::utf8($val): $val ;
-      if(defined $xml->{-pcdata}) {
-         push(@dtcontext,"-pcdata");
-         $c=$aux; 
-         $aux=&{$xml->{-pcdata}}; 
-         pop(@dtcontext);
+	?lat1::utf8($val): $val ;
+      if (defined $xml->{-pcdata}) {
+	push(@dtcontext,"-pcdata");
+	$c = $aux;
+	$aux = &{$xml->{-pcdata}};
+	pop(@dtcontext);
       }
-    }
-    else  {($atr,@val) = @$val;
-           push(@dtcontext,$name);$dtcontextcount{$name}++;
-           unshift(@dtatributes,$atr);
-           $aux = omniele($xml, $name, omni($name,$xml,@val), $atr);
-           shift(@dtatributes);
-           pop(@dtcontext);$dtcontextcount{$name}--;
+
+    } elsif (ref($tree) eq "XML::LibXML::Text") {
+#XML::PARSER#    if ($name eq "0") {
+      $val = $tree->getData();
+
+      $name = "-pcdata";
+      $aux= (defined($xml->{-outputenc}) && $xml->{-outputenc} eq 'ISO-8859-1')
+	?lat1::utf8($val): $val ;
+      if (defined $xml->{-pcdata}) {
+	push(@dtcontext,"-pcdata");
+	$c = $aux;
+	$aux = &{$xml->{-pcdata}};
+	pop(@dtcontext);
+      }
+    } else {
+      my %atr = nodeAttributes($tree);
+      $atr = \%atr;
+#XML::PARSER#      ($atr,@val) = @$val;
+
+      push(@dtcontext,$name); $dtcontextcount{$name}++;
+      unshift(@dtatributes, $atr);
+      $aux = omniele($xml, $name, omni($name, $xml, ($tree->getChildnodes())), $atr);
+#XML::PARSER#      $aux = omniele($xml, $name, omni($name,$xml,@val), $atr);
+      shift(@dtatributes);
+      pop(@dtcontext); $dtcontextcount{$name}--;
     }
     if   ($type eq "STR"){ $r .= $aux ;}
     elsif($type eq "SEQ" or $type eq "ARRAY"){
-          push(@$r,$aux) unless whitepc($aux,$name);}
+      push(@$r, $aux) unless whitepc($aux, $name);}
     elsif($type eq "SEQH" or $type eq "ARRAYHASH"){
-          push(@$r,{"-c" => $aux,
-                    "-q" => $name, 
-                    %$atr }) unless whitepc($aux,$name);}
+      push(@$r,{"-c" => $aux,
+		"-q" => $name,
+		nodeAttributes($tree) }) unless whitepc($aux,$name);}
+#XML::PARSER#                    %$atr }) unless whitepc($aux,$name);}
     elsif($type eq "MMAPON"){
-          if(not whitepc($aux,$name)){
-          if(! $typeargs{$name}) {
-              warn "duplicated tag $name\n" if(defined($r->{$name}));
-              $r->{$name}=$aux }
-          else { push(@{$r->{$name}},$aux) unless whitepc($aux,$name)}}
-          }
+      if(not whitepc($aux,$name)){
+	if(! $typeargs{$name}) {
+	  warn "duplicated tag $name\n" if(defined($r->{$name}));
+	  $r->{$name}=$aux }
+	else { push(@{$r->{$name}},$aux) unless whitepc($aux,$name)}}
+    }
     elsif($type eq "MAP" or $type eq "HASH"){
-          if(not whitepc($aux,$name)){
-          warn "duplicated tag $name\n" if(defined($r->{$name}));
-          $r->{$name}=$aux }}
+      if(not whitepc($aux,$name)){
+	warn "duplicated tag $name\n" if(defined($r->{$name}));
+	$r->{$name}=$aux }}
     elsif($type eq "MULTIMAP"){
-          push(@{$r->{$name}},$aux) unless whitepc($aux,$name)}
+      push(@{$r->{$name}},$aux) unless whitepc($aux,$name)}
     elsif($type eq "NONE"){ $r=$aux;}
     else { $r="undefined type !!!"}
   }
   $r;
-  
+
 }
 
 sub omniele {
-  my $xml=shift;
+  my $xml = shift;
   my $aux;
-  ($q,$c,$aux)=@_; 
+  ($q, $c, $aux) = @_;
+
   %v=%$aux;
+
   if (defined($xml->{-outputenc}) && $xml->{-outputenc} eq 'ISO-8859-1'){
     for (keys %v){ $v{$_} = lat1::utf8($v{$_}) ; }
   }
@@ -607,7 +690,9 @@ sub toxml {
         "$c->{-pcdata}</$q>" }
 }
 
-%mkdtskel=( '-default' => sub{$element{$q}=1; 
+sub mkdtskel{
+ my $file = shift;
+ my %mkdtskel=( '-default' => sub{$element{$q}++; 
                               for (keys %v){$att{$q}{$_}=1 }; ""},
 
              '-end' => sub{ print <<'END';
@@ -621,19 +706,40 @@ my $filename = shift;
 END
 
 for $name (keys %element){
-     print "     '$name' => sub{\"\$q:\$c\"},";
-     print '# remember $v{',
-            join('},$v{',keys %{$att{$name}}),
-            '}' if $att{$name}; 
-     print "\n";
+  print "     '$name' => sub{\"\$q:\$c\"},";
+  print '# remember $v{', join('},$v{',keys %{$att{$name}}), '}' if $att{$name};
+  print "# $element{$name}";
+  print "\n";
 }
+
 print <<'END';
 );
 print dt($filename,%handler); 
 END
-            });
+            }
+ );
 
-sub mkdtskel{ dt(shift,%mkdtskel)}
+ if($file eq "-latin1"){
+     $mkdtskel{'-inputenc'} = 'ISO-8859-1';
+     $file = shift;
+ }
+
+ dt($file,%mkdtskel)
+}
+
+sub nodeAttributes {
+  my $node = shift;
+  my %answer = ();
+  my @attrs = $node->getAttributes();
+  for (@attrs) {
+    if (ref($_) eq "XML::LibXML::Namespace") {
+      # TODO: This should not be ignored, I thing
+    } else {
+      $answer{$_->getName()} = $_->getValue();
+    }
+  }
+  return %answer;
+}
 
 sub mkdtdskel {
   my @files = @_;
@@ -647,6 +753,10 @@ sub mkdtdskel {
         },
     '-pcdata'    => sub{ if ($c =~ /[^ \t\n]/){ $ele{ctxt(1)}{"#PCDATA"}=1 }},
   );
+  if($files[0] eq "-latin1"){
+     $handler{'-inputenc'} = 'ISO-8859-1';
+     shift @files;
+  }
 
   for $filename (@files){
   dt($filename,%handler); 
@@ -664,6 +774,23 @@ sub mkdtdskel {
   }
 }
   
+
+sub process_libxml {
+  my $node = shift;
+  my $ref = ref($node);
+  if ($ref eq "XML::LibXML::Text") {
+    return [0, $node->toString() ];
+  } else {
+    my $name = $node->getName();
+    my %attr = ( map { ($_->getName(), $_->getValue()) } $node->getAttributes() );
+
+    my $x = [ $name, [ { %attr }, map {my $w =  process_libxml($_); @$w } $node->getChildnodes()  ]];
+    return $x;
+  }
+}
+
+
+
 sub putele {
   my $e = shift; 
   my @f ;
